@@ -1,4 +1,4 @@
-#     Copyright 2017, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2018, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -23,13 +23,31 @@ Exceptions from other operations are consider ErrorCodes domain.
 
 from nuitka.Options import isDebug
 
+from .CodeHelpers import generateChildExpressionsCode, generateExpressionCode
 from .ErrorCodes import getFrameVariableTypeDescriptionCode
-from .Helpers import generateChildExpressionsCode, generateExpressionCode
 from .LabelCodes import getGotoCode
-from .LineNumberCodes import emitErrorLineNumberUpdateCode
+from .LineNumberCodes import (
+    emitErrorLineNumberUpdateCode,
+    getErrorLineNumberUpdateCode
+)
 from .PythonAPICodes import getReferenceExportCode
 
 
+def generateReraiseCode(statement, emit, context):
+    context.markAsNeedsExceptionVariables()
+
+    old_source_ref = context.setCurrentSourceCodeReference(
+        value = statement.getCompatibleSourceReference()
+    )
+
+    getReRaiseExceptionCode(
+        emit    = emit,
+        context = context
+    )
+
+    context.setCurrentSourceCodeReference(old_source_ref)
+
+
 def generateRaiseCode(statement, emit, context):
     exception_type  = statement.getExceptionType()
     exception_value = statement.getExceptionValue()
@@ -73,14 +91,7 @@ def generateRaiseCode(statement, emit, context):
 
         context.setCurrentSourceCodeReference(old_source_ref)
     elif exception_type is None:
-        assert exception_cause is None
-        assert exception_value is None
-        assert exception_tb is None
-
-        getReRaiseExceptionCode(
-            emit    = emit,
-            context = context
-        )
+        assert False, statement
     elif exception_value is None and exception_tb is None:
         raise_type_name  = context.allocateTempName("raise_type")
 
@@ -130,7 +141,7 @@ def generateRaiseCode(statement, emit, context):
         getRaiseExceptionWithValueCode(
             raise_type_name  = raise_type_name,
             raise_value_name = raise_value_name,
-            implicit         = statement.isImplicit(),
+            implicit         = statement.isStatementRaiseExceptionImplicit(),
             emit             = emit,
             context          = context
         )
@@ -187,10 +198,14 @@ def generateRaiseExpressionCode(to_name, expression, emit, context):
     # Missed optimization opportunity, please report it, this should not
     # normally happen. We are supposed to propagate this upwards.
     if isDebug():
+        # TODO: Need to optimize ExpressionLocalsVariableRefORFallback once we know
+        # it handles cases where the value is not in locals dict properly.
+
         parent = expression.parent
         assert parent.isExpressionSideEffects() or \
-               parent.isExpressionConditional(), \
-               (expression, expression.parent)
+               parent.isExpressionConditional() or \
+               parent.isExpressionLocalsVariableRefORFallback(), \
+               (expression, expression.parent, expression.asXmlText())
 
     # That's how we indicate exception to the surrounding world.
     emit("%s = NULL;" % to_name)
@@ -210,7 +225,16 @@ def getReRaiseExceptionCode(emit, context):
     if keeper_variables[0] is None:
         emit(
             """\
-RERAISE_EXCEPTION( &exception_type, &exception_value, &exception_tb );"""
+%(bool_res_name)s = RERAISE_EXCEPTION( &exception_type, &exception_value, &exception_tb );
+if (unlikely( %(bool_res_name)s == false ))
+{
+    %(update_code)s
+}
+""" % {
+                "bool_res_name" : context.getBoolResName(),
+                "update_code"   : getErrorLineNumberUpdateCode(context)
+
+            }
         )
 
         frame_handle = context.getFrameHandle()
