@@ -1,4 +1,4 @@
-#     Copyright 2017, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2018, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -22,14 +22,23 @@ source code comments with developer manual sections.
 
 """
 
+from nuitka.__past__ import intern  # pylint: disable=I0021,redefined-builtin
 from nuitka.nodes.AssignNodes import (
     StatementAssignmentVariable,
     StatementReleaseVariable
 )
-from nuitka.nodes.BuiltinIteratorNodes import ExpressionBuiltinIter1
+from nuitka.nodes.AsyncgenNodes import (
+    ExpressionAsyncgenObjectBody,
+    ExpressionMakeAsyncgenObject
+)
+from nuitka.nodes.BuiltinIteratorNodes import (
+    ExpressionAsyncIter,
+    ExpressionAsyncNext,
+    ExpressionBuiltinIter1
+)
 from nuitka.nodes.BuiltinNextNodes import ExpressionBuiltinNext1
 from nuitka.nodes.CodeObjectSpecs import CodeObjectSpec
-from nuitka.nodes.ConditionalNodes import StatementConditional
+from nuitka.nodes.ConditionalNodes import makeStatementConditional
 from nuitka.nodes.ConstantRefNodes import makeConstantRefNode
 from nuitka.nodes.ContainerOperationNodes import (
     StatementListOperationAppend,
@@ -40,20 +49,18 @@ from nuitka.nodes.FrameNodes import (
     StatementsFrameFunction,
     StatementsFrameGenerator
 )
-from nuitka.nodes.FunctionNodes import (
-    ExpressionFunctionBody,
-    ExpressionFunctionCall,
-    ExpressionFunctionCreation,
-    ExpressionFunctionRef
-)
+from nuitka.nodes.FunctionNodes import ExpressionFunctionRef
 from nuitka.nodes.GeneratorNodes import (
     ExpressionGeneratorObjectBody,
-    ExpressionMakeGeneratorObject
+    ExpressionMakeGeneratorObject,
+    StatementGeneratorReturnNone
 )
 from nuitka.nodes.LoopNodes import StatementLoop, StatementLoopBreak
 from nuitka.nodes.NodeMakingHelpers import makeVariableRefNode
-from nuitka.nodes.OutlineNodes import ExpressionOutlineBody
-from nuitka.nodes.ParameterSpecs import ParameterSpec
+from nuitka.nodes.OutlineNodes import (
+    ExpressionOutlineBody,
+    ExpressionOutlineFunction
+)
 from nuitka.nodes.ReturnNodes import StatementReturn
 from nuitka.nodes.StatementNodes import (
     StatementExpressionOnly,
@@ -63,18 +70,54 @@ from nuitka.nodes.VariableRefNodes import ExpressionTempVariableRef
 from nuitka.nodes.YieldNodes import ExpressionYield
 from nuitka.PythonVersions import python_version
 
-from .Helpers import (
+from .ReformulationAssignmentStatements import buildAssignmentStatements
+from .ReformulationBooleanExpressions import buildAndNode
+from .ReformulationTryExceptStatements import makeTryExceptSingleHandlerNode
+from .ReformulationTryFinallyStatements import makeTryFinallyStatement
+from .TreeHelpers import (
     buildNode,
     buildNodeList,
+    detectFunctionBodyKind,
     getKind,
     makeStatementsSequenceFromStatement,
     makeStatementsSequenceFromStatements,
     mergeStatements
 )
-from .ReformulationAssignmentStatements import buildAssignmentStatements
-from .ReformulationBooleanExpressions import buildAndNode
-from .ReformulationTryExceptStatements import makeTryExceptSingleHandlerNode
-from .ReformulationTryFinallyStatements import makeTryFinallyStatement
+
+
+def _makeIteratorCreation(provider, qual, source_ref):
+    if getattr(qual, "is_async", 0):
+        iter_class = ExpressionAsyncIter
+    else:
+        iter_class = ExpressionBuiltinIter1
+
+    return iter_class(
+        value      = buildNode(
+            provider   = provider,
+            node       = qual.iter,
+            source_ref = source_ref
+        ),
+        source_ref = source_ref
+    )
+
+
+def _makeIteratorNext(qual, iterator_ref, source_ref):
+    if getattr(qual, "is_async", 0):
+        next_class = ExpressionAsyncNext
+    else:
+        next_class = ExpressionBuiltinNext1
+
+    return next_class(
+        value      = iterator_ref,
+        source_ref = source_ref
+    )
+
+
+def _getStopIterationName(qual):
+    if getattr(qual, "is_async", 0):
+        return "StopAsyncIteration"
+    else:
+        return "StopIteration"
 
 
 def _buildPython2ListContraction(provider, node, source_ref):
@@ -108,10 +151,7 @@ def _buildPython2ListContraction(provider, node, source_ref):
         emit_class      = StatementListOperationAppend,
         iter_tmp        = iter_tmp,
         temp_scope      = None,
-        start_value     = makeConstantRefNode(
-            constant   = [],
-            source_ref = source_ref
-        ),
+        start_value     = [],
         container_tmp   = container_tmp,
         source_ref      = source_ref,
     )
@@ -145,18 +185,18 @@ def _buildPython2ListContraction(provider, node, source_ref):
 def buildListContractionNode(provider, node, source_ref):
     # List contractions are dealt with by general code.
     if python_version < 300:
-        return _buildPython2ListContraction(provider, node, source_ref)
-
+        return _buildPython2ListContraction(
+            provider   = provider,
+            node       = node,
+            source_ref = source_ref
+        )
 
     return _buildContractionNode(
         provider    = provider,
         node        = node,
-        name        = "<listcontraction>",
+        name        = "<listcontraction>" if python_version < 370 else "<listcomp>",
         emit_class  = StatementListOperationAppend,
-        start_value = makeConstantRefNode(
-            constant   = [],
-            source_ref = source_ref
-        ),
+        start_value = [],
         source_ref  = source_ref
     )
 
@@ -169,10 +209,7 @@ def buildSetContractionNode(provider, node, source_ref):
         node        = node,
         name        = "<setcontraction>",
         emit_class  = StatementSetOperationAdd,
-        start_value = makeConstantRefNode(
-            constant   = set(),
-            source_ref = source_ref
-        ),
+        start_value = set(),
         source_ref  = source_ref
     )
 
@@ -185,10 +222,7 @@ def buildDictContractionNode(provider, node, source_ref):
         node        = node,
         name        = "<dictcontraction>",
         emit_class  = StatementDictOperationSet,
-        start_value = makeConstantRefNode(
-            constant   = {},
-            source_ref = source_ref
-        ),
+        start_value = {},
         source_ref  = source_ref
     )
 
@@ -224,37 +258,48 @@ def buildGeneratorExpressionNode(provider, node, source_ref):
         future_spec       = parent_module.getFutureSpec()
     )
 
-    code_body = ExpressionGeneratorObjectBody(
-        provider   = provider,
-        name       = "<genexpr>",
-        flags      = set(),
-        source_ref = source_ref
-    )
+    if python_version < 370:
+        is_async = any(getattr(qual, "is_async", 0) for qual in node.generators)
+    else:
+        is_async = detectFunctionBodyKind(nodes = [node])[0] in ("Asyncgen", "Coroutine")
+
+    if is_async:
+        code_body = ExpressionAsyncgenObjectBody(
+            provider   = provider,
+            name       = "<genexpr>",
+            flags      = set(),
+            source_ref = source_ref
+        )
+
+        maker_class = ExpressionMakeAsyncgenObject
+    else:
+        code_body = ExpressionGeneratorObjectBody(
+            provider   = provider,
+            name       = "<genexpr>",
+            flags      = set(),
+            source_ref = source_ref
+        )
+
+        maker_class = ExpressionMakeGeneratorObject
+
 
     function_body.setBody(
         makeStatementsSequenceFromStatements(
             StatementAssignmentVariable(
                 variable   = iter_tmp,
-                source     = ExpressionBuiltinIter1(
-                    value      = buildNode(
-                        provider   = provider,
-                        node       = node.generators[0].iter,
-                        source_ref = source_ref
-                    ),
-                    source_ref = source_ref
-                ),
+                source     = _makeIteratorCreation(provider, node.generators[0], source_ref),
                 source_ref = source_ref
             ),
             makeTryFinallyStatement(
                 provider   = function_body,
                 tried      = StatementReturn(
-                    expression = ExpressionMakeGeneratorObject(
-                        generator_ref = ExpressionFunctionRef(
+                    expression = maker_class(
+                        ExpressionFunctionRef(
                             function_body = code_body,
                             source_ref    = source_ref
                         ),
-                        code_object   = code_object,
-                        source_ref    = source_ref
+                        code_object = code_object,
+                        source_ref  = source_ref
                     ),
                     source_ref = source_ref
                 ),
@@ -280,6 +325,13 @@ def buildGeneratorExpressionNode(provider, node, source_ref):
         source_ref      = source_ref,
     )
 
+    if is_async:
+        statements.append(
+            StatementGeneratorReturnNone(
+                source_ref = source_ref
+            )
+        )
+
     statements = (
         makeTryFinallyStatement(
             provider   = function_body,
@@ -304,7 +356,8 @@ def buildGeneratorExpressionNode(provider, node, source_ref):
 
 def _buildContractionBodyNode(provider, node, emit_class, start_value,
                               container_tmp, iter_tmp, temp_scope,
-                              assign_provider, function_body, source_ref):
+                              assign_provider, function_body,
+                              source_ref):
 
     # This uses lots of variables and branches. There is no good way
     # around that, and we deal with many cases, due to having generator
@@ -314,8 +367,7 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
     # assigning one of the loop variables to the outside scope.
 
     tmp_variables = []
-
-    if assign_provider:
+    if emit_class is not ExpressionYield:
         tmp_variables.append(iter_tmp)
 
     if container_tmp is not None:
@@ -323,17 +375,12 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
 
     # First assign the iterator if we are an outline.
     if assign_provider:
+
+
         statements = [
             StatementAssignmentVariable(
                 variable   = iter_tmp,
-                source     = ExpressionBuiltinIter1(
-                    value      = buildNode(
-                        provider   = provider,
-                        node       = node.generators[0].iter,
-                        source_ref = source_ref
-                    ),
-                    source_ref = source_ref
-                ),
+                source     = _makeIteratorCreation(provider, node.generators[0], source_ref),
                 source_ref = source_ref.atInternal()
             )
         ]
@@ -344,12 +391,14 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
         statements.append(
             StatementAssignmentVariable(
                 variable   = container_tmp,
-                source     = start_value,
+                source     = makeConstantRefNode(
+                    constant   = start_value,
+                    source_ref = source_ref
+                ),
                 source_ref = source_ref.atInternal()
             )
         )
 
-
     if hasattr(node, "elt"):
         if start_value is not None:
             current_body = emit_class(
@@ -358,7 +407,7 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
                     source_ref = source_ref
                 ),
                 buildNode(
-                    provider   = function_body,
+                    provider   = function_body if not assign_provider else provider,
                     node       = node.elt,
                     source_ref = source_ref
                 ),
@@ -384,12 +433,12 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
                 source_ref = source_ref
             ),
             key        = buildNode(
-                provider   = function_body,
+                provider   = function_body if not assign_provider else provider,
                 node       = node.key,
                 source_ref = source_ref,
             ),
             value      = buildNode(
-                provider   = function_body,
+                provider   = function_body if not assign_provider else provider,
                 node       = node.value,
                 source_ref = source_ref,
             ),
@@ -425,12 +474,9 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
             nested_statements = []
         else:
             # First create the iterator and store it, next should be loop body
-            value_iterator = ExpressionBuiltinIter1(
-                value      = buildNode(
-                    provider   = function_body,
-                    node       = qual.iter,
-                    source_ref = source_ref
-                ),
+            value_iterator = _makeIteratorCreation(
+                provider   = provider if assign_provider else function_body,
+                qual       = qual,
                 source_ref = source_ref
             )
 
@@ -458,15 +504,16 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
             makeTryExceptSingleHandlerNode(
                 tried          = StatementAssignmentVariable(
                     variable   = tmp_value_variable,
-                    source     = ExpressionBuiltinNext1(
-                        value      = iterator_ref,
-                        source_ref = source_ref
+                    source     = _makeIteratorNext(
+                        iterator_ref = iterator_ref,
+                        qual         = qual,
+                        source_ref   = source_ref
                     ),
                     source_ref = source_ref
                 ),
-                exception_name = "StopIteration",
+                exception_name = _getStopIterationName(qual),
                 handler_body   = StatementLoopBreak(
-                    source_ref = source_ref.atInternal()
+                    source_ref = source_ref
                 ),
                 source_ref     = source_ref
             ),
@@ -483,21 +530,19 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
         ]
 
         conditions = buildNodeList(
-            provider   = function_body,
+            provider   = provider if assign_provider else function_body,
             nodes      = qual.ifs,
             source_ref = source_ref
         )
 
         if len(conditions) >= 1:
             loop_statements.append(
-                StatementConditional(
+                makeStatementConditional(
                     condition  = buildAndNode(
                         values     = conditions,
                         source_ref = source_ref
                     ),
-                    yes_branch = makeStatementsSequenceFromStatement(
-                        statement = current_body
-                    ),
+                    yes_branch = current_body,
                     no_branch  = None,
                     source_ref = source_ref
                 )
@@ -549,46 +594,20 @@ def _buildContractionNode(provider, node, name, emit_class, start_value,
     # described in the developer manual. They use a lot of temporary names,
     # nested blocks, etc. and so a lot of variable names.
 
-    # TODO: No function ought to be necessary, only a variable scope is.
-    function_body = ExpressionFunctionBody(
+    function_body = ExpressionOutlineFunction(
         provider   = provider,
-        name       = name,
-        doc        = None,
-        parameters = ParameterSpec(
-            ps_name          = name,
-            ps_normal_args   = (".0",),
-            ps_list_star_arg = None,
-            ps_dict_star_arg = None,
-            ps_default_count = 0,
-            ps_kw_only_args  = ()
-        ),
-        flags      = set(),
+        name       = intern(name[1:-1]),
         source_ref = source_ref
     )
 
-    iter_tmp = function_body.getVariableForAssignment(
-        variable_name = ".0"
-    )
-    assert iter_tmp.isParameterVariable()
-
-    parent_module = provider.getParentModule()
-
-    code_object = CodeObjectSpec(
-        co_name           = name,
-        co_kind           = "Function",
-        co_varnames       = function_body.getParameters().getParameterNames(),
-        co_argcount       = len(function_body.getParameters().getParameterNames()),
-        co_kwonlyargcount = 0,
-        co_has_starlist   = False,
-        co_has_stardict   = False,
-        co_filename       = parent_module.getRunTimeFilename(),
-        co_lineno         = source_ref.getLineNumber(),
-        future_spec       = parent_module.getFutureSpec()
+    iter_tmp = function_body.allocateTempVariable(
+        temp_scope = None,
+        name       = ".0"
     )
 
     container_tmp = function_body.allocateTempVariable(
         temp_scope = None,
-        name       = "contraction_result"
+        name       = "contraction"
     )
 
     statements, release_statements = _buildContractionBodyNode(
@@ -604,7 +623,11 @@ def _buildContractionNode(provider, node, name, emit_class, start_value,
         source_ref      = source_ref,
     )
 
-    assert start_value
+    assign_iter_statement = StatementAssignmentVariable(
+        source     = _makeIteratorCreation(provider, node.generators[0], source_ref),
+        variable   = iter_tmp,
+        source_ref = source_ref
+    )
 
     statements.append(
         StatementReturn(
@@ -625,37 +648,36 @@ def _buildContractionNode(provider, node, name, emit_class, start_value,
         ),
     )
 
-    function_body.setBody(
-        makeStatementsSequenceFromStatement(
-            statement = StatementsFrameFunction(
+    if python_version < 300:
+        body = makeStatementsSequenceFromStatements(
+            assign_iter_statement,
+            statements
+        )
+    else:
+        parent_module = provider.getParentModule()
+
+        code_object = CodeObjectSpec(
+            co_name           = name,
+            co_kind           = "Function",
+            co_varnames       = (),
+            co_argcount       = 1,
+            co_kwonlyargcount = 0,
+            co_has_starlist   = False,
+            co_has_stardict   = False,
+            co_filename       = parent_module.getRunTimeFilename(),
+            co_lineno         = source_ref.getLineNumber(),
+            future_spec       = parent_module.getFutureSpec()
+        )
+
+        body = makeStatementsSequenceFromStatements(
+            assign_iter_statement,
+            StatementsFrameFunction(
                 statements  = mergeStatements(statements, False),
                 code_object = code_object,
                 source_ref  = source_ref
             )
         )
-    )
 
-    return ExpressionFunctionCall(
-        function   = ExpressionFunctionCreation(
-            function_ref = ExpressionFunctionRef(
-                function_body = function_body,
-                source_ref    = source_ref
-            ),
-            code_object  = code_object,
-            defaults     = (),
-            kw_defaults  = None,
-            annotations  = None,
-            source_ref   = source_ref
-        ),
-        values     = (
-            ExpressionBuiltinIter1(
-                value      = buildNode(
-                    provider   = provider,
-                    node       = node.generators[0].iter,
-                    source_ref = source_ref
-                ),
-                source_ref = source_ref
-            ),
-        ),
-        source_ref = source_ref
-    )
+    function_body.setBody(body)
+
+    return function_body
